#!/usr/bin/python3


################################################################################
# Imports
################################################################################

import tkinter
from tkinter.filedialog import askopenfilename

import sys
import sqlite_utils as db

from multiprocessing import Process, JoinableQueue

################################################################################
# Constantes
################################################################################

MAX_COL_SIZE = 30
SQLITE_KEYWORDS = [
    'ABORT', 'ACTION', 'ADD', 'AFTER', 'ALL', 'ALTER', 'ANALYZE', 'AND', 'AS',
    'ASC', 'ATTACH', 'AUTOINCREMENT', 'AVG', 'BEFORE', 'BEGIN', 'BETWEEN', 'BY',
    'CASCADE', 'CASE', 'CAST', 'CHECK', 'COLLATE', 'COLUMN', 'COMMIT',
    'CONFLICT', 'CONSTRAINT', 'COUNT', 'CREATE', 'CROSS', 'CURRENT_DATE',
    'CURRENT_TIME', 'CURRENT_TIMESTAMP', 'DATABASE', 'DEFAULT', 'DEFERRABLE',
    'DEFERRED', 'DELETE', 'DESC', 'DETACH', 'DISTINCT', 'DROP', 'EACH', 'ELSE',
    'END', 'ESCAPE', 'EXCEPT', 'EXCLUSIVE', 'EXISTS', 'EXPLAIN', 'FAIL', 'FOR',
    'FOREIGN', 'FROM', 'FULL', 'GLOB', 'GROUP', 'HAVING', 'IF', 'IGNORE',
    'IMMEDIATE', 'IN', 'INDEX', 'INDEXED', 'INITIALLY', 'INNER', 'INSERT',
    'INSTEAD', 'INTERSECT', 'INTO', 'IS', 'ISNULL', 'JOIN', 'KEY', 'LEFT',
    'LIKE', 'LIMIT', 'MIN', 'MAX', 'MATCH', 'NATURAL', 'NO', 'NOT', 'NOTNULL',
    'NULL', 'OF', 'OFFSET', 'ON', 'OR', 'ORDER', 'OUTER', 'PLAN', 'PRAGMA',
    'PRIMARY', 'QUERY', 'RAISE', 'RECURSIVE', 'REFERENCES', 'REGEXP', 'REINDEX',
    'RELEASE', 'RENAME', 'REPLACE', 'RESTRICT', 'RIGHT', 'ROLLBACK', 'ROW',
    'SAVEPOINT', 'SELECT', 'SET', 'SUM', 'TABLE', 'TEMP', 'TEMPORARY', 'THEN', 'TO',
    'TRANSACTION', 'TRIGGER', 'UNION', 'UNIQUE', 'UPDATE', 'USING', 'VACUUM',
    'VALUES', 'VIEW', 'VIRTUAL', 'WHEN', 'WHERE', 'WITH', 'WITHOUT']
SQLITE_STRINGS = [("'", "'"), ('"', '"')]
SQLITE_COMMENTS = [("--", "\n")]
SQLITE_BRACKETS = ["(", ")", "[", "]", "{", "}"]
SQLITE_WARNINGS = [';(.|\n)*[^ \n]']
KEYWORD_COLOR = "#1493E2"
STRING_COLOR = "#FC9A24"
COMMENT_COLOR = "#BBBBBB"
BRACKET_COLOR = "#444444"
WARNING_COLOR = "#B02020"

TIMEREACT = 500  # reaction time to end of queries (in microsecond)

################################################################################
# Fonctions utilitaires
################################################################################

def column_size(tab):
    sizes = None
    for r in tab:
        if sizes == None:
            sizes = [0] * len(r)
        n = 0
        for v in r:            
            sizes[n] = max(sizes[n], len(str(v)))
            n += 1
            
    return sizes

def expand(s, n):
    if n < len(s):
        return s[0:n-6] + " [...] "
    return s + (" " * (n - len(s)))


################################################################################
# Un widget de texte avec coloration syntaxique
################################################################################

# Une classe permettant d'équiper le widget d'un callback
# réagissant aux modifications
# (code honteusement repris de Stackoverflow)

class ModifiedMixin:
    def _init(self):
        self.clearModifiedFlag()
        self.bind_all('<<Modified>>', self._beenModified)

    def _beenModified(self, event=None):
        if self._resetting_modified_flag: return
        self.clearModifiedFlag()
        self.beenModified(event)

    def beenModified(self, event=None):
        pass

    def clearModifiedFlag(self):
        self._resetting_modified_flag = True

        try:
            self.tk.call(self._w, 'edit', 'modified', 0)

        finally:
            self._resetting_modified_flag = False           


class CodeText(tkinter.Text, ModifiedMixin):   
    def __init__(self, *args, **kwargs):
        tkinter.Text.__init__(self, *args, **kwargs)
        self._init()

        self.keywords = []
        self.intervals = []
        self.highlighting = 1

    def beenModified(self, event=None):
        self.tag_remove("keyword", "1.0", "end")
        self.tag_remove("warning", "1.0", "end")
        if self.highlighting:
            self.highlight_patterns(self.keywords, 1)
            self.highlight_patterns(self.brackets, 0, False)
            self.highlight_intervals(self.intervals)
            self.highlight_patterns(self.warnings)

    def highlight_patterns(self, patterns, nocase=0, regexp=True, start="1.0", end="end"):
        for p in patterns:
            self.highlight_pattern(p[0], p[1], nocase, regexp, start, end)

    def highlight_pattern(self, pattern, tag, nocase=0, regexp=True, start="1.0", end="end"):

        start = self.index(start)
        end = self.index(end)
        self.mark_set("matchStart", start)
        self.mark_set("matchEnd", start)
        self.mark_set("searchLimit", end)

        count = tkinter.IntVar()
        while True:
            index = self.search(pattern, "matchEnd","searchLimit",
                                count=count, regexp=regexp, nocase=nocase)
            if index == "": break
            self.mark_set("matchStart", index)
            self.mark_set("matchEnd", "%s+%sc" % (index, count.get()))
            self.tag_add(tag, "matchStart", "matchEnd")

    def highlight_intervals(self, patterns, regexp=False, start="1.0", end="end"):                          

        for p in patterns:
            self.tag_remove(p[2], "1.0", "end")


        start = self.index(start)
        end = self.index(end)
        self.mark_set("matchStart", start)
        self.mark_set("matchEnd", start)
        self.mark_set("searchLimit", end)

        count = tkinter.IntVar()
        while True:
            selected_pattern = None
            selected_index = end
            for p in patterns:
                index = self.search(p[0], "matchEnd","searchLimit",
                                    count=count, regexp=regexp)
                if (index != "" and self.compare(index, '<', selected_index)):
                    selected_index = index
                    selected_pattern = p
                
            if selected_pattern == None: break
            self.mark_set("matchStart", "%s" % (selected_index))
            self.mark_set("searchStart", "%s+1c" % (selected_index))
            index = self.search(selected_pattern[1], "searchStart","searchLimit",
                                count=count, regexp=regexp)
            if index == "":
                index = "%s-2c" % (end)
            self.tag_add(selected_pattern[2], selected_index, "%s+1c" % (index))
            self.mark_set("matchEnd", "%s+1c" % (index))

    def select_all(self):
        self.tag_add(tkinter.SEL, "1.0", tkinter.END)
        self.mark_set(tkinter.INSERT, "1.0")
        self.see(tkinter.INSERT)

################################################################################
# Fenêtres
################################################################################

class SQLWindow(tkinter.Tk):    
    def __init__(self, dbfile=''):
        tkinter.Tk.__init__(self)
        self.wm_title("Exécuteur de requêtes SQL")
        self.protocol("WM_DELETE_WINDOW", self.exit)
        container = tkinter.Frame(self)
        container.pack(side="top", fill="both", expand=True)

        menu = tkinter.Menu(self)
        self.config(menu=menu)
        filemenu = tkinter.Menu(menu)
        menu.add_cascade(label="Fichier", menu=filemenu)
        filemenu.add_command(label="Ouvrir un fichier DB...", command=self.choose_file, accelerator='Ctrl+O')
        filemenu.add_command(label="Exécuter/Stopper la requête", command=self.run_query, accelerator='Ctrl+E')
        filemenu.add_separator()
        filemenu.add_command(label="Quitter", command=self.exit, accelerator='Ctrl+Q')

        editmenu = tkinter.Menu(menu)
        menu.add_cascade(label="Édition", menu=editmenu)
        editmenu.add_command(label="Sélectionner tout", command=self.select_all, accelerator='Ctrl+A')

        self.bind_all('<Control-o>', self.choose_file)
        self.bind_all('<Control-e>', self.run_query)
        self.bind_all('<Control-q>', self.exit)        
        self.bind_all('<Control-a>', self.select_all)        

        canvas = tkinter.Canvas(container)
        tkinter.Label(canvas, text="Code SQL à exécuter :", justify=tkinter.LEFT).pack(fill='y', side='left')
        self.highlighting = tkinter.IntVar()
        self.highlighting.set(1)
        check = tkinter.Checkbutton(canvas, text="Coloration syntaxique", variable=self.highlighting)
        check.configure(command=self.toggle_highlighting)
        check.pack(fill='y', side='right')
        canvas.pack(fill='x')
        
        self.create_sqltextview(container)

        self.button_execute = tkinter.Button(container, text="Exécuter la requête", command=self.run_query)
        self.button_execute.pack()

        canvas2 = tkinter.Canvas(container)
        tkinter.Label(canvas2, text="Retour du Système de Gestion de Bases de Données :", justify=tkinter.LEFT).pack(fill='y', side='left')
        tkinter.Label(canvas2, text="").pack(fill='y', side='right')
        canvas2.pack(fill='x')

        self.create_answertextview(container)

        self.running_query = None # Thread of query 
        self.running_queue = None # queue for query result

        self.connection_status = tkinter.Label("")
        self.connection_status.pack(fill='x')
        self.dbfile = dbfile
        self.update_status()

    def toggle_highlighting(self):
        self.sqltextview.highlighting = self.highlighting.get()
        self.sqltextview.beenModified()

    def create_sqltextview(self, container):
        canvas = tkinter.Canvas(container)
        vscrollbar = tkinter.Scrollbar(canvas)
        self.sqltextview = CodeText(canvas, height=10, width=100)

        # On ajoute les styles
        self.sqltextview.tag_configure("warning", foreground=WARNING_COLOR)
        self.sqltextview.tag_configure("keyword", foreground=KEYWORD_COLOR)
        self.sqltextview.tag_configure("string", foreground=STRING_COLOR)
        self.sqltextview.tag_configure("comment", foreground=COMMENT_COLOR)
        self.sqltextview.tag_configure("bracket", foreground=BRACKET_COLOR)

        # On ajoute les règles spécifiques SQL
        self.sqltextview.keywords = [('( |\n|\t|^|\(|\))' + x + '( |\n|\t|$|\(|\))', 'keyword') for x in SQLITE_KEYWORDS]
        self.sqltextview.brackets = [(x, 'bracket') for x in SQLITE_BRACKETS]
        self.sqltextview.intervals = ([(x[0], x[1], 'string') for x in SQLITE_STRINGS] +
                                      [(x[0], x[1], 'comment') for x in SQLITE_COMMENTS])
        self.sqltextview.warnings = [(x, 'warning') for x in SQLITE_WARNINGS]

        vscrollbar.config(command=self.sqltextview.yview)
        self.sqltextview.pack(side='left', fill='both', expand=True)
        vscrollbar.pack(side='right', fill='y')
        canvas.pack(fill='both', expand=True)

        self.sqltextview.configure(yscrollcommand=vscrollbar.set)
        self.sqltextview.configure(font=('monospace'))

    def create_answertextview(self, container):
        canvas = tkinter.Canvas(container)
        vscrollbar = tkinter.Scrollbar(canvas)
        hscrollbar = tkinter.Scrollbar(canvas)
        self.answertextview = tkinter.Text(canvas, height=20, width=100, state=tkinter.DISABLED, wrap=tkinter.NONE)
        canvas.columnconfigure(0, weight=1)
        canvas.rowconfigure(0, weight=1)
        
        vscrollbar.config(command=self.answertextview.yview)
        hscrollbar.config(command=self.answertextview.xview, orient=tkinter.HORIZONTAL)
        self.answertextview.grid(row=0, column=0, sticky=tkinter.N+tkinter.S+tkinter.E+tkinter.W)
        vscrollbar.grid(row=0, column=1, sticky=tkinter.N+tkinter.S)
        hscrollbar.grid(row=1, column=0, sticky=tkinter.E+tkinter.W)
        canvas.pack(fill='both', expand=True)
        self.answertextview.configure(yscrollcommand=vscrollbar.set)
        self.answertextview.configure(xscrollcommand=hscrollbar.set)
        self.answertextview.configure(font=('monospace'))
        self.answertextview.tag_configure('error', foreground='#ff0000')
        self.answertextview.tag_configure('normal', foreground='#000000')
        self.answertextview.tag_configure('warning', foreground='#A0A0A0')

    def select_all(self, event = None):
        self.sqltextview.select_all()
        
    def run_query(self, event=None):
        if not self.dbfile:
            self.display_error("""Aucune connexion active...

Connectez-vous d'abord en utilisant "Ouvrir un fichier DB" dans le menu "Fichier".""")
            return
        if not self.running_query is None:
            self.close_running_query()
            return
        self.display_warning("""Exécution de la requête en cours ...""")
        self.button_execute.configure(text="Stopper la requête")        
        query = self.sqltextview.get(1.0, tkinter.END)
        self.running_queue = JoinableQueue()
        self.running_query = Process(target=db.execute_query,
                                     args=(self.dbfile, query, self.running_queue))
        self.running_query.start()
        self.after(TIMEREACT, self.check_running_query, event)
        
    def close_running_query(self):
        sys.stderr.flush()
        if self.running_queue:
            self.running_queue.close()
            self.running_queue = None
        if self.running_query and self.running_query.is_alive():
            print("fin de requête")
            self.display_error("""Requête interrompue""")
            self.running_query.terminate()
        sys.stdout.flush()
        self.running_query = None
        self.button_execute.configure(text="Exécuter la requête")
        
    def check_running_query(self, event=None):
        if self.running_query is None:
            return
        if not self.running_query.is_alive():
            self.display_error("Fin de la requête inattendue")
            self.close_running_query()
            return
        try:
            answer = self.running_queue.get_nowait()
        except:
            self.after(TIMEREACT, self.check_running_query, event)
            return
        self.close_running_query()
        if answer[2] != None: # Une erreur...
            s = "Erreur d'exécution de la requête...\n"
            s += "=" * 80
            s += "\n\n"
            s += answer[2]
            self.display_error(s)
            return

        if len(answer[1]) == 0: # Aucune ligne retournée par la requête...
            s = "Aucune ligne retournée par la requête\n"
            s += "=" * 80
            s += "\n\n"
        else:
            s = "{0} ligne{1} retournée{1} par la requête\n".format(len(answer[1]), '' if len(answer[1]) < 2 else 's')
            s += "=" * 80
            s += "\n\n"

            sizes = column_size(answer[1])
            for i in range(0, len(sizes)):
                sizes[i] = min(max(sizes[i], len(answer[0][i][0])), MAX_COL_SIZE)

            n = 0
            for v in answer[0]:
                s += expand(str(v[0]), sizes[n] + 2)
                n += 1
            s += '\n'
            s += '-' * (2 * len(sizes) + sum(sizes))
            s += '\n'

            for r in answer[1]:
                n = 0
                for v in r:
                    s += expand(str(v), sizes[n] + 2)
                    n += 1
                s += '\n'

        self.display_answer(s)

    def display_answer(self, s):
        self.answertextview.configure(state=tkinter.NORMAL)
        self.answertextview.delete(1.0, tkinter.END)
        self.answertextview.insert(tkinter.INSERT, s, 'normal')
        self.answertextview.configure(state=tkinter.DISABLED)

    def display_error(self, s):
        self.answertextview.configure(state=tkinter.NORMAL)
        self.answertextview.delete(1.0, tkinter.END)
        self.answertextview.insert(tkinter.INSERT, s, 'error')
        self.answertextview.configure(state=tkinter.DISABLED)

    def display_warning(self, s):
        self.answertextview.configure(state=tkinter.NORMAL)
        self.answertextview.delete(1.0, tkinter.END)
        self.answertextview.insert(tkinter.INSERT, s, 'warning')
        self.answertextview.configure(state=tkinter.DISABLED)

    def exit(self, event=None):
        self.close_running_query()
        self.destroy()

    def choose_file(self, event=None):
        name = askopenfilename(
            title="Fichier de base de données SQLite",
            filetypes=[('Fichiers Sqlite', '*.db'), ('Tous les fichiers', '*')]
            )

        if name:
            self.dbfile = name
            self.update_status()


    def update_status(self):
        if not self.dbfile:
            self.connection_status.configure(text="(Non connecté)")
        else:
            self.connection_status.configure(text="Connecté au fichier {0}".format(self.dbfile))

            
if __name__ == "__main__":   
    sql_win = SQLWindow(sys.argv[1] if len(sys.argv) > 1 else '')
    # sqlWin.connect("delete-event", tkinter.main_quit)
    sql_win.mainloop()
