import sqlite3
import traceback
import os, sys
try:
    import resource
except:
    print("Impossible de charger le paquet 'resource'.")
    print("Ce n'est pas grave, mais la limitation en RAM ne sera pas possible.")

connection = None

def set_connection(fn):
    def wrapped(dbfile, *args, **kwargs):
        global connection
        if not connection:
            init_connection(dbfile)
        return fn(connection, *args, **kwargs)
    return wrapped

OOM_HINT = ("Vérifiez que vous n'avez pas oublié une condition dans une jointure.\n"
            "Si le problème persite, essayez de relancer l'exécuteur de requêtes.\n")
   
LIMIT_MEMORY_DONE = False
def limit_memory_usage():
    global LIMIT_MEMORY_DONE
    if LIMIT_MEMORY_DONE:
        return
    try:
        # Si le code échoue plus loin, on ne re-tentera pas.
        LIMIT_MEMORY_DONE = True
        # En théorie, RSS (resident size) ou DATA devrait suffire, mais ça n'a
        # pas l'air de marcher d'après mes tests. Solution paranoiaque en
        # limitant tout ce qu'on peut en espérant que l'un marche. D'après mes
        # tests, c'est AS (addres space) qui marche.
        for r in (resource.RLIMIT_AS, resource.RLIMIT_DATA, resource.RLIMIT_RSS):
            try:
                # 1GB of memory ought to be enough for anybody
                resource.setrlimit(r, (1.5e9, resource.RLIM_INFINITY))
            except ValueError:
                print("Failed to set memory limit. Continue at your own risk")
    except NameError:
        pass
    except:
        print("Un problème est survenu en essayant de mettre en place la limite de RAM")
        print("Ce n'est pas grave, on continue sans limitation")

#########################################################
# Cette fonction exécute sur la base
# la requête passée en paramètre
#########################################################
@set_connection
def execute_query(connection, query, queue):
    limit_memory_usage()
    try:
        # On récupère un objet curseur qui permet de parcourir l'ensemble résultat à la manière d'un itérateur.
        cursor = connection.cursor()
        
        # On exécute la requête ici.
        cursor.execute(query)

        # Après exécution de la requête, on récupère la réponse du SGBD et on renvoie le tout.
        answer = (cursor.description, cursor.fetchall(), None)

    except sqlite3.Warning as e:
        # Se produit s'il y a un ';' dans la requete (ailleurs qu'à la fin)
        msg = e.args[0]
        answer = (None, None, msg)
        
    except MemoryError:
        # Tester avec par exemple SELECT a1.* FROM Artistes AS a1, Artistes AS a2;
        print(traceback.format_exc())
        answer = (None, None, "Pas assez de mémoire pour exécuter la requête SQL.\n" + OOM_HINT)

    except sqlite3.Error as e:
        print(traceback.format_exc())
        if len(e.args) > 0:
            msg = e.args[0]
        else:
            # Il arrive qu'on obtienne sqlite3.OperationalError sur un
            # "out of memory", et quand c'est le cas, e.args est un
            # tableau vide donc le e.args[0] lancerait une exception
            # dans le traitant d'exception :-(. On devine le diagnostique.
            msg = ("Erreur pendant l'exécution de la requête.\n" + 
                   "Cette erreur peut se produire s'il n'y a pas assez de mémoire.\n" + OOM_HINT)
        answer = (None, None, msg)
        
    except:
        print(traceback.format_exc())
        answer = (None, None, "Exception inattendue.\n")
    
    sys.stdout.flush()
    sys.stderr.flush()
    close_connection()
    queue.put(answer)
    queue.join()
    
#########################################################
# Cette fonction exécute sur la base
# la requête de mise-à-jour passée en paramètre
#########################################################
@set_connection
def execute_update(connection, query):
    try:
        # On récupère un objet curseur qui permet de parcourir l'ensemble résultat à la manière d'un itérateur.
        cursor = connection.cursor()
        
        # On exécute la requête ici.
        cursor.execute(query)

    except sqlite3.Error as e:
        print("Erreur d'exécution de la requête - %s:" % e.args[0])
    

def commit():
    global connection
    
    if not connection:
        sqlite3.initConnection()

    try:
        connection.commit()

    except sqlite3.Error as e:
        print("Erreur d'exécution de la requête - %s:" % e.args[0])
    

#########################################################
# Cette fonction initialise la connexion à la base
#########################################################
def init_connection(dbfile):
    global connection
    
    if not os.path.isfile(dbfile):
        print("Impossible d'ouvrir la base de données %s:\n"
              "le fichier n'existe pas." %
              dbfile)
        sys.exit(1)

    try:
        connection = sqlite3.connect(dbfile)
        # patch to "correct" MAX function
        # connection.create_aggregate("MAX", 1, MyMax)
        
    except sqlite3.Error as e:
        print("Database connexion error - %s:" % e.args[0])
        close_connection()


#########################################################
# Cette fonction ferme la connexion à la base
#########################################################
def close_connection():
    global connection
    
    if connection:
        connection.close()
        connection = None
        

# patch to correct MAX aggregation function
class MyMax:
    def __init__(self):
        self.max = None

    def step(self, value):
        try:
            v = int(value)
            if self.max is None:
                self.max = v
                return
            self.max = max(self.max, v)
        except:
            pass

    def finalize(self):
        return self.max

